# %% [markdown]
# # Circuits with Bloqade
#
# In this tutorial we will demonstrate how to write circuits and quantum executions with Bloqade. Specifically, we will use the `squin` dialect set from the compiler toolchain `kirin`. `SQUIN` stands for `S`tructural `Qu`antum `IN`struction set and is the circuit-level representation of quantum executions. It is built on top of the `kirin` framework, an [open-source compiler toolchain](https://queracomputing.github.io/kirin/latest/) for embedded domain-specific languages (eDSLs) that target scientific computing kernels. A key feature of squin is the _kernel_, which can roughly be seen as the object which will be executed on the target hardware. Naturally, this hardware could be a quantum computer, but it also extends to classical execution as well, such as mid-circuit feedforward or even non-quantum execution such as robotics.
#
# These kernels can be built using decorators of python functions. We will use the `@squin.kernel` decorator in this notebook but keep in mind that other eDSLs have different decorators inherited from base Kirin decorators. The decorator lowers Python's abstract syntax tree (AST) into a kirin SSA (single static assignment) form, which is a useful intermediate representation for compiler analysis. You don't have to worry too much about SSA or compilers here, but if you want to learn more check out the [kirin documentation](https://queracomputing.github.io/kirin/latest/).

# %%
from typing import Any

import numpy as np
import bloqade.types
from kirin.ir import Method
from bloqade.types import Qubit, MeasurementResult

# Some types we will use, useful for type hints
from kirin.dialects.ilist import IList

from bloqade import squin

Register = IList[Qubit, Any]


# %%
@squin.kernel
def hello_world(theta: float) -> IList[MeasurementResult, Any]:
    """
    Prepare a Bell state and measure in a basis that might have a Bell violation
    """
    qubits = squin.qalloc(2)
    squin.h(qubits[0])
    squin.cx(qubits[0], qubits[1])
    squin.rx(theta, qubits[0])
    bits = squin.qubit.measure(qubits)
    return bits


# [kernel].print() prints the raw SSA, which is the intermediate representation of the kernel
# as used internally by Kirin.
hello_world.print()
# %% [markdown]
# ### Anatomy of a Bloqade kernel
#
# A kernel is a representation of a hybrid quantum/classical execution that will run "on hardware". The decorator @squin.kernel can be considered as the "label" to represent this fact. A kernel can take arguments (such as `theta` here) and can return values (such as `bits` here). To help the compiler, the inputs and outputs must be decorated with a type. The return values can be considered as the "results" which are returned from the quantum computer, though they can also be intermediate values that are used as a part of a larger computation (e.g. a function call). A kernel can represent both quantum and classical execution: while lots of classical computation is supported (control flow, algebra, etc.) there is less support for arbitrary python calls, as this won't necessarily be supported on the microcontroller "on hardware". Note that shot-level execution and averaging is handled outside of the kernel itself: averaging over multiple shots within a kernel is an anti-pattern to avoid.

# %% [markdown]
# ## Squin kernel statements
#
# To represent quantum executions, we write and construct squin kernels.
# These kernels can use the typical structure of python functions -- inputs, outputs, for loops, control flow, subroutines, and so forth -- as a feature of the underlying base Kirin statements.
# Because one can intermix control flow with operations and measurement, arbitrary mid-circuit feed forward comes for "free".
#
# There are three dialects that comprise the domain-specific language
# 1. `squin.qubit` - Manipulation and declaration of the qubits themselves, mainly allocating new ones and measuring them.
# 2. `squin.gate` - Application of quantum gates on qubits.
# 3. `squin.noise` - Representing noise effects on the qubits.
#
# While you can interact with `squin.qubit` directly, the other two dialects are exposed via wrapper functions that are available directly under the `squin` namespace, or under `squin.broadcast` for parallelized versions of the gates and noise processes, respectively.
#
# Refer to the [API documentation](../../../reference/bloqade-circuit/src/bloqade/squin/stdlib/simple/gate) for a full list of the available functionality.
# Note that you are also able to define your own custom gates by defining similar functions.


# %% [markdown]
# # Using Bloqade kernels
# A key feature of kernels is the ability to do complex control flow similar to how one might program python. For example, one can use a for loop to apply the same gate to multiple qubits to prepare a GHZ state.
#
# A useful pattern is to use factory functions that return bloqade kernels. This way, you can fix parameters (such as the number of qubits) pythonically without needing to introduce the variable directly into the kernel itself, using a closure.

# %%
# Bell state prep.


def GHZ_method_factory(nqubits: int) -> Method:
    @squin.kernel
    def GHZ_state() -> Register:
        qubits = squin.qalloc(nqubits)
        squin.h(qubits[0])
        for i in range(nqubits):
            squin.cx(qubits[i], qubits[i + 1])

        return qubits

    return GHZ_state


kernel = GHZ_method_factory(8)
kernel.print()
# %% [markdown]
# Alternatively, kernels could be parameterized; for example, we could write the same GHZ state preparation, except it prepares a variable number of qubits that is not declared until the kernel is run. In order to run in some `main` function, the qubits need to be declared elsewhere, either in the task declaration or within a larger kernel that calls this method as a subroutine.


# %%
@squin.kernel
def GHZ_state_factory(nqubits: int) -> Register:
    qubits = squin.qalloc(nqubits)
    squin.h(qubits[0])
    for i in range(nqubits - 1):
        squin.cx(qubits[i], qubits[i + 1])
    return qubits


GHZ_state_factory.print()
# %% [markdown]
# ## Building circuits in Cirq
# Instead of writing your circuit directly in bloqade, you may build circuits using Cirq and then lower them to and from bloqade kernels. This has the advantage of being able to leverage the excellent and in-depth resources of transpilation and circuit optimization without having to reinvent the wheel. However, for certain programs, such as those requiring more complex mid-circuit feed-forward, it is still required to write bloqade kernels as there is no adequate representation in other SDKs. Cirq is our initial choice of SDK, and other transformations are coming soon-- though in principle interoperability with many SDK is possible through an intermediate Cirq representation.
#
# Let us begin by writing a simple GHZ state preparation circuit, in analogy to the bloqade kernel above. Observe that the resulting object is a static representation of a circuit, similar to the `GHZ_state` kernel, and differentiated from the dynamic `GHZ_state_factory` kernel which can return a dynamically sized GHZ state.

# %%
import cirq


def ghz_prep(nqubits: int) -> cirq.Circuit:
    """
    Builder function that returns a simple N-qubit
    GHZ state preparation circuit
    """
    qubits = cirq.LineQubit.range(nqubits)
    output = cirq.Circuit()
    output.append(cirq.H(qubits[0]))
    for i in range(nqubits - 1):
        output.append(cirq.CX(qubits[i], qubits[i + 1]))
    return output


print(ghz_prep(4))
# %% [markdown]
# The cirq circuit can be converted to a bloqade kernel with the transpilation function `load_circuit`. The kernel can be considered as a transformation on the register of qubits it is applied to as arguments, with the return being the qubits that still persist.

# %%
from bloqade.cirq_utils import emit_circuit, load_circuit

# Load a cirq circuit into squin
kernel = load_circuit(
    ghz_prep(4),
    kernel_name="ghz_prep_cirq",  # Define the name of the kernel as if one were using @squin.kernel on a function
    register_as_argument=False,  # If the resulting kernel should take in a qubit register (True) or make a new one (False)
    return_register=True,  # If the resulting kernel should return the register of the qubits it acts on.
)

# Then, we can convert the circuit back to cirq.
# Note that this is **not possible** in a general case because
# cirq cannot represent complex control flow.
circuit2: cirq.Circuit = emit_circuit(kernel, ignore_returns=True)
print(circuit2)
# %% [markdown]
# The circuit loading also works with classical feed forward, though it is generally more difficult to extract a cirq circuit from a generic feedforward cirq kernel. For example, the T teleportation gadget can be written and loaded as

# %%
reg = cirq.LineQubit.range(2)
circuit = cirq.Circuit()
circuit.append(cirq.T(reg[0]))
circuit.append(cirq.CNOT(reg[1], reg[0]))
circuit.append(cirq.measure(reg[0], key="m"))
circuit.append(cirq.S(reg[1]).with_classical_controls("m"))
circuit.append(cirq.X(reg[1]).with_classical_controls("m"))
print(circuit)
kernel = load_circuit(
    circuit, kernel_name="teleport", register_as_argument=True, return_register=True
)
kernel.print()
# %% [markdown]
# Due to the difficulty of representing mid-circuit control flow in cirq, attempting to lower these kernels back to cirq will result in an error.


# %%
@squin.kernel
def t_teleport_noargs() -> None:
    """
    A simple T teleportation circuit that requires mid circuit control flow.
    """
    ancilla = squin.qalloc(1)[0]
    target = squin.qalloc(1)[0]
    squin.t(ancilla)
    squin.cx(target, ancilla)
    if squin.qubit.measure(target):
        squin.s(ancilla)
        squin.x(ancilla)


try:
    print(emit_circuit(t_teleport_noargs))
    raise (RuntimeError("Oops this should have errored."))
except Exception as e:
    print("ERROR:", e)


# Though measurement without feedforward is possible
@squin.kernel
def coinflip() -> MeasurementResult:
    qubit = squin.qalloc(1)[0]
    squin.h(qubit)
    return squin.qubit.measure(qubit)


circuit = emit_circuit(coinflip, ignore_returns=True)
print(circuit)
# %% [markdown]
# ## Simulation, emulation, and analysis
#
# A kernel is simply a representation of an execution and is not much use without being able to analyze and execute that kernel. We can simulate the action of kernels using concrete interpreters. The emulator must a) keep track of the classical state of the variables, b) keep track of the quantum state of the qubits, and thus c) faithfully represent the execution of that program as if it was run on a hybrid quantum/classical computer. Bloqade's emulator is built on top of the excellent [PyQrack quantum simulator](https://pyqrack.readthedocs.io/en/latest/) and satisfies the three goals above.
#
# There are four main objects when considering simulation:
# 1. **The emulator object** - Representing the thing that some kernel is going to be executed on. Today it is the PyQrack simulator, but eventually it could also include other simulators or physical hardware. The `*.task` method of the `emulator` object builds a `task` object.
# 2. **The task object** - Binding together the emulator, kernel, and input parameters for that kernel. This task is not executed until the `*.run` method is called. Upon calling `run`, the kernel is interpreted, the quantum circuit is executed, any classical co-processing is done, and the kernel completes, returning . Repeated calling of `run` will "reset" the executor into its initial state and rerun the kernel. Alternatively, one could call `*.batch_run` to repeatedly run the kernel's result to get stochastic averaging. Products of the task include the `results` and `QuantumState` objects.
# 3. **The results object** - Whatever the `return` of the kernel is returned, with the same type signature. This is generated from `*run()` (as a ResultType) or `*.batch_run` (as a dict keyed by ResultType and valued by frequency) These could be qubits or qubit registers (list of qubits), values, or whatever object you like.
# 4. **The QuantumState object** - The final quantum state of the emulator object. While this is a nonphysical quantity, the QuantumState is useful for debugging and analysis. This can be extracted from either the `emulator.quantum_state` method (for a single run after the `run()` method), or with `task.batch_state` (for a stochastic average over many samples). The quantum state is efficiently represented as an eigensystem of a reduced density matrix.

# %%
from bloqade.pyqrack import StackMemorySimulator, DynamicMemorySimulator

# StackMemorySimulator - static number of qubits
# DynamicMemorySimulator - dynamic number of qubits, but slower. Use if you don't know how many qubits you need in advance.
emulator = StackMemorySimulator(min_qubits=8)

task = emulator.task(GHZ_state_factory, args=(4,))
results = task.run()
# The results are the same ResultType as the kernel return.
# In this case, it is a list of qubits.
print(results)
# %% [markdown]
# Note that, while it is instinctive to simply call `GHZ_state_factory(4)` and expect it to run, this isn't necessarily the correct abstraction. `emulator.task` links the kernel to a specific interpreter-- for example, if you wanted to run your program on a noisy emulator vs a perfect emulator vs. real hardware, this is the way you would specify it. Furthermore, the _instantiation_ of the task is not necessarily linked with the _execution_ of that task. For example, if that task must be run asynchronously on hardware, there must be an object which represents the run itself as well as the (future) results.
#
# For simpler kernels that only use kirin built-in statements -- such as control flow, for loops, arithmetic, and the like -- it is possible to directly call the kernel and use the default Kirin interpreter.


# %%
def foo(x: int, y: int) -> bool:
    return x < y


assert foo(1, 2)
# %% [markdown]
# ## Extracting quantum states
#
# PyQRack is mainly intended as an *emulator*, attempting to recreate the physical action and interaction of quantum hardware with classical controls. However, it is often useful for user analysis and debugging to extract out non-physical values, such as the quantum statevector, fidelity, or exact expectation values of observables and correlation functions.
#
# These nonphysical values are based in being able to directly observe and evaluate the quantum state. The state can be extracted via the `PyQrackSimulator.quantum_state` method, with an input signature of a list of (pyqrack) qubits. These qubits can be generated as a return value from a kernel (as is the case for the `GHZ_state_factory` function) or from the `task.qubits()` method.
#
# The resulting state is a reduced density matrix represented by its eigensystem of eigenvalues and eigenvectors, or a dense $2^N \times 2^N$ array.
# %%
from bloqade.pyqrack import PyQrackQubit

# This kernel returns a list of Qubit objects. We can use these to analyze the
# quantum state of the register.
results: list[PyQrackQubit] = task.run()

# If the kernel does not return the qubit register, one can still collect it. Note that
# it is now the onus of you, the user, to determine which qubits are which.
# Qubits are typically added sequentially, so if you make multiple registers, the qubits
# will be in the order they were added. The StackMemorySimulator may have extra qubits.
qubits: list[PyQrackQubit] = task.qubits()

# Extract the quantum state as a reduced density matrix. Note that he qubits themselves
# point to the internal state of the emulator
state = emulator.quantum_state(results)
density_matrix = emulator.reduced_density_matrix(results)

# Note that the RDM is represented in its eigenbasis for efficiency.
# If the state is pure, there is only one nonzero eigenvalue. This is the case for the GHZ state.
print("A pure state has only a single eigenvalue [1] of the RDM:", state.eigenvalues)
statevector = state.eigenvectors[:, 0]
print("The statevector of the GHZ state looks like [1, 0, 0, ..., 0, 1]")
print(statevector)
# %% [markdown]
# If the output is randomized, one can average over many runs using `task.batch_run`. This returns a dictionary of probabilities of each output. Note that the output must be hashable.

# %%
# Define the emulator and task
emulator = StackMemorySimulator(min_qubits=1)
task = emulator.task(coinflip)
results = task.batch_run(shots=1000)
state = task.batch_state(shots=1000)
print("Results:", results)
print("State:", state)


# %% [markdown]
# # Composition of kernels
#
# Bloqade kernels allow all the typical syntax of for loops, if-else statements, function calls, and other powerful abstractions. Let us use this to write efficient representations of complex circuits.
#
# For this example, we will use a [Trotterization of the 1d Transverse Ising model](https://qiskit-community.github.io/qiskit-algorithms/tutorials/13_trotterQRTE.html).
#
# The first option we will explore is to write the entire circuit in Cirq and then convert it into a bloqade kernel using the `load_circuit` lowering. Observe that the return objects of these builder functions are static objects.
# %%
def trotter_layer(
    qubits: list[cirq.Qid], dt: float = 0.01, J: float = 1, h: float = 1
) -> cirq.Circuit:
    """
    Cirq builder function that returns a circuit of
    a Trotter step of the 1D transverse Ising model
    """
    op_zz = cirq.ZZ ** (dt * J / np.pi)
    op_x = cirq.X ** (dt * h / np.pi)
    circuit = cirq.Circuit()
    for i in range(0, len(qubits) - 1, 2):
        circuit.append(op_zz.on(qubits[i], qubits[i + 1]))
    for i in range(1, len(qubits) - 1, 2):
        circuit.append(op_zz.on(qubits[i], qubits[i + 1]))
    for i in range(len(qubits)):
        circuit.append(op_x.on(qubits[i]))
    return circuit


def trotter_circuit(
    N: int, steps: int = 10, dt: float = 0.01, J: float = 1, h: float = 1
) -> cirq.Circuit:
    qubits = cirq.LineQubit.range(N)
    circuit = cirq.Circuit()
    for _ in range(steps):
        circuit += trotter_layer(qubits, dt, J, h)
    return circuit


cirq_trotter_circuit = trotter_circuit(N=8, steps=4, dt=0.01, J=1, h=1)

print(cirq_trotter_circuit)

# Convert the circuit to a bloqade kernel
bloqade_trotter_circuit = load_circuit(
    cirq_trotter_circuit,
    kernel_name="trotter",
    register_as_argument=False,
    return_register=True,
)
# %% [markdown]
# As an intermediate, one can mix between writing kernels converted from Cirq circuits and direct bloqade kernels. For example, each layer has fixed parameters as defined by a cirq circuit, but a variable number of layers as parameterized by a kernel input and for loop. This option has the benefit of being able to use Cirq infrastructure to optimize and represent individual layers, while still being able to use bloqade kernels to represent parameterized circuits. In this case, the output kernel has the timestep and Ising parameters fixed (as they are fixed in the cirq circuit), but the number of steps is variable.


# %%
def factory_trotter(N: int, dt: float = 0.01, J: float = 1, h: float = 1) -> Method:
    bloqade_trotter_layer = load_circuit(
        trotter_layer(qubits=cirq.LineQubit.range(N), dt=dt, J=J, h=h),
        kernel_name="trotter",
        register_as_argument=True,
        return_register=True,
    )

    @squin.kernel
    def trotter_for_loop(steps: int) -> Register:
        """
        Main function that runs the Trotter circuit for a given number of steps
        """
        qubits = squin.qalloc(N)
        for _ in range(steps):
            qubits = bloqade_trotter_layer(qubits)
        return qubits

    return trotter_for_loop


# %% [markdown]
# Alternatively, you could just write everything directly as a Bloqade kernel. Note that the ZZ operator that is native to Cirq must be expanded into its own "helper" kernel via a decomposition into cx / z / cx. The resulting kernel is fully parameterized, with the values not actually evaluated until runtime (or further compilation and folding).
# %%
# Define an operator that looks like the ZZ power gate
@squin.kernel
def op_zz(theta: float, qb1: bloqade.types.Qubit, qb2: bloqade.types.Qubit) -> None:
    """
    A kernel that returns an operator that looks like ZZ^{theta/2pi}
    """
    squin.cx(qb1, qb2)
    squin.rz(theta, qb2)
    squin.cx(qb1, qb2)


@squin.kernel
def bloqade_trotter(
    N: int, steps: int, dt: float = 0.01, J: float = 1, h: float = 1
) -> Register:
    """
    Main function that runs the Trotter circuit for a given number of steps
    """
    qubits = squin.qalloc(N)
    for _ in range(steps):
        for i in range(0, len(qubits) - 1):
            op_zz(theta=dt * J, qb1=qubits[i], qb2=qubits[i + 1])
        for i in range(0, len(qubits)):
            squin.rx(angle=dt * h, qubit=qubits[i])
    return qubits


# %% [markdown]
# Of course, both Cirq and the (converted) Bloqade kernel have the same execution and same output state.

# %%
cirq_trotter = trotter_circuit(N=12, steps=10, dt=0.01, J=1, h=1)

cirq_statevector = cirq.Simulator().simulate(cirq_trotter).state_vector()

# Or converting to a bloqade kernel and simulating with PyQrack
cirq_trotter_kernel = load_circuit(
    cirq_trotter,
    kernel_name="cirq_trotter",
    register_as_argument=False,
    return_register=True,
)
cirq_trotter_qubits = (
    StackMemorySimulator(min_qubits=12).task(cirq_trotter_kernel).run()
)
cirq_bloqade_state = StackMemorySimulator.quantum_state(cirq_trotter_qubits)
# The state is, of course, the same. A little bit of work is needed to extract out the (single) state vector from the RDM.
print(
    "Overlap:",
    np.abs(np.dot(np.conj(cirq_statevector), cirq_bloqade_state.eigenvectors[:, 0]))
    ** 2,
)
# %% [markdown]
# Similarly, the execution and output state of the cirq state and kernel written fully in bloqade have the same state. Note that for the `bloqade_trotter` kernel, the arguments are not declared until the simulator is run.

# %%
cirq_trotter_qubits = StackMemorySimulator(min_qubits=12).run(
    bloqade_trotter, args=(12, 10, 0.01, 1, 1)
)
cirq_bloqade_state = StackMemorySimulator.quantum_state(cirq_trotter_qubits)
print(
    "Overlap:",
    np.abs(np.dot(np.conj(cirq_statevector), cirq_bloqade_state.eigenvectors[:, 0]))
    ** 2,
)


# %% [markdown]
# # Mid-circuit feed forward
#
# Bloqade kernels can natively represent mid-circuit feed-forward using control flow represented by standard pythonic if-else and while structures. While the possibilities are endless, including measurement-based quantum computing and error correction, we show two examples here.
#
# The first is T state teleportation, which teleports a T gate that was applied to an ancilla (a "T state") onto the target state using only Clifford gates and feedforward. Due to the property of being Clifford, the circuit itself is fault tolerant and thus plays an important role in many error corrected algorithms.
# %%
@squin.kernel
def t_teleport(target: squin.qubit.Qubit) -> squin.qubit.Qubit:
    ancilla = squin.qalloc(1)[0]
    squin.h(ancilla)
    squin.t(ancilla)
    squin.cx(control=target, target=ancilla)
    bit = squin.qubit.measure(target)
    if bit:
        squin.s(ancilla)
    return ancilla  # The state of the target qubit is also teleported to the ancilla


# And now let’s wrap it into a larger context to run. In this case,
# apply to a |+> state and see that we get a T|+> state out.
@squin.kernel
def t_teleport_wrapper() -> squin.qubit.Qubit:

    target = squin.qalloc(1)[0]
    squin.h(target)
    target = t_teleport(target)
    return target


# And run it. Observe that the batch_state uses a qubit_map to select which qubits to include in the batch state.
# This is important because there are two qubits total (the target and the ancilla) but we only want inspect
# the state of the output qubit.
emulator = StackMemorySimulator(min_qubits=2)
task = emulator.task(t_teleport_wrapper)
state = task.batch_state(shots=1000, qubit_map=lambda x: [x])
# Even though there is measurement and feedforward, the final state is still pure. Neat!
print(state)
# %% [markdown]
# ### Constant depth GHZ state
#
# Remarkably, it is also possible to prepare a GHZ state with a constant number of gates. At first glance, this seems impossible: quantum information can only propagate as fast as the information lightcone, and so a constant depth circuit can only have a small amount of entanglement—not enough to prepare the long-range correlated GHZ state. The trick is, we can replace quantum gates with ancillas and classical feedforward measurements so that the information is propagated classically instead of quantumly. For more details, check out [1] and [2].
#
# [1] [Efficient Long-Range Entanglement Using Dynamic Circuits](https://doi.org/10.1103/PRXQuantum.5.030339)
#
# [2] [Constant-Depth Preparation of Matrix Product States with Adaptive Quantum Circuits](https://doi.org/10.1103/PRXQuantum.5.030344)
#
# The explicit circuit for the GHZ circuit is shown in Fig. 5 of [1]. There is classical feedforward in the form of a parity check, which requires a classical XOR operation that is irrepresentable by CIRQ.


# %%
def ghz_constant_depth(n_qubits: int):

    @squin.kernel
    def main() -> Register:
        qreg = squin.qalloc(n_qubits)
        ancilla = squin.qalloc(n_qubits - 1)

        for i in range(n_qubits):
            squin.h(qreg[i])

        for i in range(n_qubits - 1):
            squin.cx(qreg[i], ancilla[i])
        for i in range(n_qubits - 1):
            squin.cx(qreg[i + 1], ancilla[i])

        parity: int = 0
        bits = squin.broadcast.measure(ancilla)
        for i in range(n_qubits - 1):
            parity = parity ^ bits[i]
            if parity == 1:
                squin.x(qreg[i + 1])
        return qreg

    return main


# %%
# At this point, you know the drill. We can simulate this with multirun via PyQrack
emulator = StackMemorySimulator(min_qubits=7)
task = emulator.task(ghz_constant_depth(3))

state = task.batch_state(shots=1000, qubit_map=lambda x: x)
# Even though there is measurement and feedforward, the final state is still pure. Neat!
print(state.eigenvalues)
# %% [markdown]
# As a final note, consider how difficult it would be to represent this circuit in Cirq. In particular, there is a for loop, where inside the for loop there is an algebraic operation (XOR) that feeds forward onto a variable (parity). This circuit is very hard to express in Cirq without some serious hacking of ancilla registers.



"""Generate the code reference pages and navigation."""

import os
from pathlib import Path

import mkdocs_gen_files

if os.getenv("GITHUB_ACTIONS") == "true":
    BLOQADE_CIRCUIT_SRC_PATH = "submodules/bloqade-circuit/"
    BLOQADE_ANALOG_SRC_PATH = "submodules/bloqade-analog/"
else:
    """
    NOTE: we assume the following project structure when building locally:

    ../
    ├── bloqade
    ├── bloqade-analog
    └── bloqade-circuit
    """
    BLOQADE_CIRCUIT_SRC_PATH = "../bloqade-circuit/"
    BLOQADE_ANALOG_SRC_PATH = "../bloqade-analog/"


skip_keywords = [
    "julia",  ## [KHW] skip for now since we didn't have julia codegen rdy
    "builder/base",  ## hiding from user
    "builder/terminate",  ## hiding from user
    "ir/tree_print",  ## hiding from user
    "ir/visitor",  ## hiding from user
    "codegen/",  ## hiding from user
    "builder/factory",  ## hiding from user
    "builder_old",  ## deprecated from user
    "task_old",  ## deprecated from user
    "visualization",  ## hiding from user
    "submission/capabilities",  ## hiding from user
    "submission/quera_api_client",
    "test/",
    "tests/",
    "test_utils",
    "docs/",
    "debug/",
    "squin/cirq/emit/",  # NOTE: this fails when included because there is an __init__.py missing, but the files have no docs anyway and it will be moved so safe to ignore
]


def make_nav(bloqade_package_name: str, BLOQADE_PACKAGE_PATH: str):
    """
    build the mkdocstrings nav object for the given package

    Arguments:
        bloqade_package_name (str): name of the bloqade package. This must match with the mkdocs path as the generated pages are put under reference/<bloqade_package_name>
        BLOQADE_PACKAGE_PATH (str): the path to the module.
    """
    nav = mkdocs_gen_files.Nav()
    for path in sorted(Path(BLOQADE_PACKAGE_PATH).rglob("*.py")):
        module_path = Path(path.relative_to(BLOQADE_PACKAGE_PATH).with_suffix(""))
        doc_path = Path(
            bloqade_package_name, module_path.relative_to(".").with_suffix(".md")
        )
        full_doc_path = Path("reference/", doc_path)

        iskip = False

        for kwrd in skip_keywords:
            if kwrd in str(doc_path):
                iskip = True
                break
        if iskip:
            print("[Ignore]", str(doc_path))
            continue

        print("[>]", str(doc_path))

        parts = tuple(module_path.parts)

        if parts[-1] == "__init__":
            parts = parts[:-1]
            doc_path = doc_path.with_name("index.md")
            full_doc_path = full_doc_path.with_name("index.md")
        elif parts[-1].startswith("_"):
            continue

        if len(parts) == 0:
            continue

        nav[parts] = doc_path.as_posix()
        with mkdocs_gen_files.open(full_doc_path, "w") as fd:
            ident = ".".join(parts[1:])
            fd.write(f"::: {ident}")

        mkdocs_gen_files.set_edit_path(full_doc_path, ".." / path)

    return nav


bloqade_circuit_nav = make_nav("bloqade-circuit", BLOQADE_CIRCUIT_SRC_PATH)
with mkdocs_gen_files.open("reference/SUMMARY_BLOQADE_CIRCUIT.md", "w") as nav_file:
    nav_file.writelines(bloqade_circuit_nav.build_literate_nav())

bloqade_analog_nav = make_nav("bloqade-analog", BLOQADE_ANALOG_SRC_PATH)
with mkdocs_gen_files.open("reference/SUMMARY_BLOQADE_ANALOG.md", "w") as nav_file:
    nav_file.writelines(bloqade_analog_nav.build_literate_nav())



# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -all
#     custom_cell_magics: kql
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.17.2
#   kernelspec:
#     display_name: kirin-workspace (3.12.10)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # GHZ State Preparation with Squin
#
# In this example, we will show (yet again) how to implement a program that prepares a GHZ state.
# We will do so with a simple linear algorithm and show how to manually insert noise.

# %% [markdown]
# The circuit we will implement is displayed below:
#
# ![GHZ linear circuit](../../ghz_linear_circuit.svg)

# %% [markdown]
# Let's start by importing Squin and writing our circuit for an arbitrary number of qubits.

# %%
from bloqade.pyqrack import StackMemorySimulator  # we'll need that later

from bloqade import squin


@squin.kernel
def ghz_linear(n: int):
    q = squin.qalloc(n)
    squin.h(q[0])
    for i in range(1, n):
        squin.cx(q[i - 1], q[i])


ghz_linear.print()

# %% [markdown]
# As you can see, writing basic circuits in squin is rather straightforward.


# %% [markdown]
# ## Simulating the kernel
#
# You can simulate a kernel such as the above using bloqade's PyQrack backend.
#
# There are two basic simulators, that act like "devices" that you run your program on:
#
# * The `StackMemorySimulator`, which initializes its memory with a fixed number of qubits. The number is either set via the `min_qubits` argument or inferred automatically. Note, that automatic inference is not always possible in which case you will be required to set the argument accordingly.
# * The `DynamicMemorySimulator`, which, as the name suggests, allocates memory as required throughout the circuit. Generally, you should prefer the `StackMemorySimulator` over this one unless the number of qubits can only be known at runtime.
#
# Let's go ahead and use the `StackMemorySimulator` together with a fixed number of qubits to simulate our GHZ preparation program from above.

# %%
sim = StackMemorySimulator(min_qubits=2)
result = sim.run(ghz_linear, args=(2,))
print(result)

# %% [markdown]
# As you can see, the result of our simulation is `None`.
# That is because we are not returning anything from the kernel function.
#
# Note, how we're passing in the arguments of the kernel function as a separate tuple in the call to `run`.
# This signature is required since `ghz_linear(2)` would not return another kernel function, but rather attempt to run the kernel function as a Python function.
# As the function is written in squin rather than Python, this would fail.
# To provide a little more detail here: the `PyQrack` backend in bloqade-circuit actually has its own method table which tells it how to interpret the statements encountered in the squin kernel function.
#
# Since we are only simulating the circuit, however, we are able to fetch information that would otherwise not be attainable.
# For example, you can obtain the state vector from the simulator:

# %%
print(sim.state_vector(ghz_linear, args=(2,)))


# %% [markdown]
# Looking at the output, we can see that we indeed prepared a two-qubit GHZ state (up to a global phase).
#
# Note, that you can also add a return value to the kernel, which is then returned by `sim.run`.
# Again, this is not generally possible when running on hardware, but only during simulation.
#
# A realistic kernel function will return (a list of) measurement results.
# That is precisely what we will do in the following.

# %% [markdown]
# ## Inserting noise
#
# The above is rather basic, so let's try to do something that is a little more interesting.
# Let's write the same program as before, but now we assume that noise processes occur whenever a gate is applied.
#
# We will make use of Squin's noise submodule in order to do that.
#
# Our "noise model" will be quite simple:
# * Whenever a single-qubit gate is applied, that qubit undergoes a depolarization error with probability `p_single`.
# * Whenever a two-qubit (controlled) gate is applied, both qubits undergo a joint depolarization error with probability `p_paired`.
#
# Note, that a depolarization error with probability $p$ on a single qubit means that randomly chosen Pauli operators (one of $X, Y, Z$) is applied to the qubit with probability $p$.
# Similarly, a two-qubit depolarization error applies one of the 15 operators $IX, IY, IZ, XI, XX, ...$ with a given probability.


# %%
@squin.kernel
def noisy_linear_ghz(n: int, p_single: float, p_paired: float):
    q = squin.qalloc(n)

    squin.h(q[0])
    squin.depolarize(p_single, q[0])

    for i in range(1, n):
        squin.cx(q[i - 1], q[i])
        squin.depolarize2(p_paired, q[i - 1], q[i])

    return squin.broadcast.measure(q)


# %%
noisy_linear_ghz.print()

# %% [markdown]
# <div class="admonition note">
# <p class="admonition-title">Noise operators</p>
# <p>
#     As opposed to standard gates, there is no standard library for noise statements as of now.
#     While we plan to add that in the future, also note how it can be convenient to separate the operator
#     from the gate application: we define the paired noise operator only once and apply it to different
#     pairs of qubits in the loop.
# </p>
# </div>

# %% [markdown]
# This kernel function can be simulated in the exact same way as before.
# The only difference is that we now need to provide additional arguments for the noise probabilities.

# %%
result = sim.run(noisy_linear_ghz, args=(2, 1e-2, 2e-2))
print(result)

# %% [markdown]
# Now that we actually return something, we also obtain a result from running the simulation.
# This result is just a list of measurement results (boolean values corresponding to 0 and 1).
# We can also obtain the bit string:

# %%
result_bitstring = [int(res) for res in result]
print(result_bitstring)

# %% [markdown]
# Ideally, the two values should always be correlated since we want to prepare a GHZ state.
# However, now that we've added noise, this is not always the case.
#
# We can actually use this fact to define a "fidelity" measure for the circuit: when repeatedly executing the circuit, uncorrelated results lower the fidelity.
#
# Mathematically, let's define the fidelity $F$ as
#
# $F = 1 - \sum_{i=1}^n \frac{\text{err}(i)}{n}$,
#
# where $n$ is the number of shots we take and
#
# $ \text{err}(i) = \begin{cases} 0 \text{ if run }i \text{ is correct} \\ 1 \text{ else } \end{cases}$
#
# In this case, "correct" means the measurement outcome is fully correlated.

# %%
n_shots = 1000
n_qubits = 4
p_single = 1e-2
p_paired = 2 * p_single
fidelity = 1.0
sim = StackMemorySimulator(min_qubits=n_qubits)
for _ in range(n_shots):
    result = sim.run(noisy_linear_ghz, args=(n_qubits, p_single, p_paired))
    measured_one_state = all(result)
    measured_zero_state = not any(result)
    is_correlated = measured_one_state or measured_zero_state
    if not is_correlated:
        fidelity -= 1 / n_shots

print(fidelity)

# %% [markdown]
# Note, that this is actually a poor measure for fidelity as it only counts fully correlated states and treats everything else as an equivalent error.
# If you have many qubits, you could argue that only flipping a single bit is a much lower error than flipping many, and that this should be weighed in here.
# Or, you can simply use the simulator to obtain the state vector and compute the overlap.
# Or, define whatever measure of fidelity you see fit here, but we'll end this tutorial here.



# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -all
#     custom_cell_magics: kql
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.17.2
#   kernelspec:
#     display_name: kirin-workspace (3.12.10)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # GHZ State preparation and noise
#
# In this example, we will illustrate how to work with `bloqade`'s heuristic noise models of Gemini class digital quantum processors by applying them to a circuit that prepares a GHZ state.

# %% [markdown]
# ## Primer on Gemini noise models
#
# In `bloqade`, there are two classes of heuristic noise models: one-zone models such as `GeminiOneZoneNoiseModel` and a two-zone model `GeminiTwoZoneNoiseModel`.
# These are inspired by two distinct approaches to implement a quantum circuit on hardware and are designed to get a sense of the influence of noise on Gemini class hardware.
#
# On the one hand, the one-zone model assumes a single-zone layout where qubits remain in the gate zone throughout the computation.
# On the other hand, the two-zone model incorporates a storage zone and assumes that qubits are transported between gate and storage regions.
#
# Both models are informed by benchmark data on the device but are intentionally conservative.
# Specifically, they tend to overestimate noise due to the lack of knowledge about optimized move schedules, which leads to overestimating move-induced errors.
#
# At this stage, we recommend interpreting the two models as providing a range for expected noise levels on Gemini-class devices, rather than precise predictions. They are useful for gaining intuition about noise sensitivity and for benchmarking algorithmic robustness to errors, using hardware-informed but simplistic assumptions.
#
# Note that there are actually two additional one-zone noise models, `GeminiOneZoneNoiseModelCorrelated` and `GeminiOneZoneNoiseModelConflictGraphMoves`.
# As the names suggest, the former also takes into account correlated noise, whereas the latter takes into account more realistic move schedules.
# In the following example, we will not be considering these two, but they are interchangeable with the used noise models (up to the fact, that the conflict graph moves require you to specify qubits as `cirq.GridQubit`s).

# %% [markdown]
# ## Noise model implementations
#
# For now, these noise models are implemented as [`cirq.NoiseModel`](https://quantumai.google/reference/python/cirq/NoiseModel) classes, so that you can use with any circuit you build using `cirq`.
# They are part of the [`bloqade.cirq_utils`](../../../../reference/bloqade-circuit/src/bloqade/cirq_utils) submodule.
#
# Support for using these models with e.g. [squin](../../../dialects_and_kernels) will follow in the future.
# However, you can already rely on [interoperability with cirq](../../../cirq_interop) in order to convert (noisy) circuits to squin kernels and use other parts of the compiler pipeline.

# %% [markdown]
# ## GHZ preparation and noise
#
# Now, let's get started with the actual example.
#
# As a first step, we will define a function that builds a GHZ circuit in cirq that has a depth linear in the number of qubits.
#

# %%
import cirq
import numpy as np
import matplotlib.pyplot as plt
from bloqade.cirq_utils import noise, transpile, load_circuit


def ghz_circuit(n: int) -> cirq.Circuit:
    qubits = cirq.LineQubit.range(n)

    # Step 1: Hadamard on the first qubit
    circuit = cirq.Circuit(cirq.H(qubits[0]))

    # Step 2: CNOT chain from qubit i to i+1
    for i in range(n - 1):
        circuit.append(cirq.CNOT(qubits[i], qubits[i + 1]))

    return circuit


# %% [markdown]
# ### Closer look at a basic circuit

# %% [markdown]
# Here's what this circuit looks like for `n=3` qubits:

# %%
ghz_circuit_3 = ghz_circuit(3)
print(ghz_circuit_3)

# %% [markdown]
# So far so good.
# Now, we will convert the circuit above to a noisy one using bloqade's `cirq_utils` submodule.
#
# Specifically, we can use the `noise.transform_circuit` utility function with a noise model of our choice.

# %%
noise_model = noise.GeminiOneZoneNoiseModel()
noisy_ghz_circuit_3 = noise.transform_circuit(ghz_circuit_3, model=noise_model)
print(noisy_ghz_circuit_3)

# %% [markdown]
# As you can see, we have successfully added noise.
# However, the circuit also looks very different in terms of its gates.
#
# This is because `noise.transform_circuit` does actually two things:
#
# 1. Since we want to consider a circuit that is compatible with the Gemini architecture, we need to transform it to the native gate set first. This set consists of (phased) X gates and CZ gates only.
# 2. Once we have a native circuit, noise is injected according to the chosen noise model.
#
# To clarify, here is how you would convert the circuit without using the `noise.transform_circuit` utility function:

# %%
native_ghz_3 = transpile(ghz_circuit_3)
print(native_ghz_3)

# %% [markdown]
# Note that `transpile` basically just wraps cirq's own `cirq.optimize_for_target_gateset(circuit, gateset=cirq.CZTargetGateset())`, with some additional benefits (such as filtering out empty moments).
#
# Using this native circuit, we can obtain the same noisy circuit as before by simply using cirq's `cirq.Circuit.with_noise` method.

# %%
noisy_ghz_circuit_3 = native_ghz_3.with_noise(noise_model)
print(noisy_ghz_circuit_3)

# %% [markdown]
# ### Studying the fidelity
#
# Now that we have got the basics down, we can compute the fidelity of noisy circuits with different qubit numbers.
# By fidelity, we simply mean the overlap of the final state with the perfect GHZ state expected from the noise-less version of the circuit.
#
# The corresponding density matrices are obtained using `cirq`'s simulator.
#
# We will do the simulation using two different noise models, the one-zone model used above and also the two-zone model.

# %% [markdown]
# <div class="admonition note">
# <p class="admonition-title">Fidelity calculation</p>
# <p>
#     In the following, we will simply use the expectation value of the noisy density matrix computed against the noiseless one as a proxy for fidelity.
#     This is a suboptimal choice, but we wanted to keep the example simple.
#     Feel free to substitute the fidelity calculation by the fidelity of your choice (e.g. the Uhlmann fidelity)
# </p>
# </div>

# %%
qubits = range(3, 9)

one_zone_model = noise.GeminiOneZoneNoiseModel()
two_zone_model = noise.GeminiTwoZoneNoiseModel()
simulator = cirq.DensityMatrixSimulator()

fidelities_one_zone = []
fidelities_two_zone = []
for n in qubits:
    circuit = ghz_circuit(n)
    one_zone_circuit = noise.transform_circuit(circuit, model=one_zone_model)
    two_zone_circuit = noise.transform_circuit(circuit, model=two_zone_model)

    rho = simulator.simulate(circuit).final_density_matrix
    rho_one_zone = simulator.simulate(one_zone_circuit).final_density_matrix
    rho_two_zone = simulator.simulate(two_zone_circuit).final_density_matrix

    fidelity_one_zone = np.trace(rho @ rho_one_zone)
    fidelity_two_zone = np.trace(rho @ rho_two_zone)

    fidelities_one_zone.append(fidelity_one_zone)
    fidelities_two_zone.append(fidelity_two_zone)

# %% [markdown]
# Now, let's have a look at the results.

# %%
plt.plot(qubits, fidelities_one_zone, "o", label="one-zone model")
plt.plot(qubits, fidelities_two_zone, "x", label="two-zone model")
plt.xlabel("Number of qubits")
plt.ylabel("Fidelity")
plt.legend()

# %% [markdown]
# <div align="center">
# <picture>
#    <img src="../noisy_ghz_fidelity.svg" >
# </picture>
# </div>

# %% [markdown]
# We can see that in both cases the fidelity goes down when increasing the number of qubits.
#
# Interestingly, there is a cross-over point where the two-zone model starts to exhibit a better fidelity.
# This is because as the number of qubits grows, the error introduced on idle qubits inside the gate zone is larger in the one-zone model since all qubits are always inside the gate zone.
# Whereas, in the two-zone model, qubits are moved between the gate and storage zones.
#
# You could now think about how to optimize the circuits in order to reduce their sensitivity to noise.
# For example, you can [reduce the circuit depth](../../qasm2/ghz)

# %% [markdown]
# ### Modifying the noise
#
# There are a number of parameters that govern the effect a noise model introduces into a circuit.
# These can all be set independently to adapt the noise model to your specific application.
#
# In general, there are noise parameters for the following noise processes:
#
# * Depolarization due to gate application.
# * Depolarization due to movement, both applied to moving atoms and idle atoms (a.k.a. sitter errors).
# * Atom loss errors.
#
# <div class="admonition note">
# <p class="admonition-title">Atom loss</p>
# <p>
#     Please note that atom loss is currently not supported, i.e. it's not considered in the noise models.
#     We plan to add that in the future.
# </p>
# </div>
#
# The noise processes are further split into local and global noise channels and separated by their cause.
#
# For a full list of noise parameters and a description of each one, please refer to the move noise model in [`bloqade.qasm2.dialects.noise.model.MoveNoiseModelABC`](../../../../reference/bloqade-circuit/src/bloqade/qasm2/dialects/noise/model#MoveNoiseModelABC)
#
# We can use those parameters in order to modify the strength of the noise.
#
# For example, say you want to introduce an extra penalty for moving qubits around in order to study how you can reduce movements. To do so, let's re-use the fidelity calculation using the two-zone model from above, but modify movement errors.
# We can query the default move errors from the `cirq` noise model:

# %%
default_model = noise.GeminiTwoZoneNoiseModel()
px, py, pz = default_model.mover_px, default_model.mover_py, default_model.mover_pz
print(
    f"The noise Pauli channel associated with moving atoms is (px, py, pz) = ({px,py,pz})."
)

# %% [markdown]
#
# Then we can instantiate a noise model with modified parameters:
# %%
modified_two_zone_model = noise.GeminiTwoZoneNoiseModel(
    mover_px=2e-3,
    mover_py=2e-3,
    mover_pz=3e-3,
)
fidelities_modified_two_zone = []
for n in qubits:
    circuit = ghz_circuit(n)
    noisy_circuit = noise.transform_circuit(circuit, model=modified_two_zone_model)
    rho = simulator.simulate(circuit).final_density_matrix
    rho_noisy = simulator.simulate(noisy_circuit).final_density_matrix
    fidelities_modified_two_zone.append(np.trace(rho @ rho_noisy))

# %%
plt.plot(qubits, fidelities_one_zone, "o", label="one-zone model")
plt.plot(qubits, fidelities_modified_two_zone, "x", label="modified two-zone model")
plt.xlabel("Number of qubits")
plt.ylabel("Fidelity")
plt.legend()

# %% [markdown]
# <div align="center">
# <picture>
#    <img src="../noisy_ghz_modified.svg" >
# </picture>
# </div>

# %% [markdown]
# As you can see, the fidelities no longer cross over since the increased movement noise now eliminates the advantage of the two-zone model for the considered numbers of qubits.

# %% [markdown]
# ## Interoperability with squin
#
# Finally, we want to point out that you can also use the generated noisy circuits to obtain a squin kernel function.
#
# This is useful if you want to use other features of the bloqade pipeline.
# For example, it would allow you to run the `pyqrack` simulator instead of `cirq`'s own, which can be more efficient.

# %%
circuit = ghz_circuit(5)
noisy_circuit = noise.transform_circuit(circuit, model=noise.GeminiOneZoneNoiseModel())

# %%
kernel = load_circuit(circuit, kernel_name="kernel")
noisy_kernel = load_circuit(noisy_circuit, kernel_name="noisy_kernel")
kernel.print()



# %% [markdown]
# # Deutsch-Jozsa Algorithm
# In this example, we will implement a version of the [Deutsch-Josza algorithm](https://en.wikipedia.org/wiki/Deutsch–Jozsa_algorithm) using bloqade's squin dialect.
# %% [markdown]

# We start by loading in some stuff and defining some parameters.

# %%
import random
from typing import Any

from bloqade.types import Qubit
from kirin.dialects import ilist
from bloqade.pyqrack import StackMemorySimulator

from bloqade import squin

n_bits = 2

# %% [markdown]
#
# Now, before we can implement the actual algorithm, we need to define the oracles, i.e. the functions we want to check for.
#
# The problem is defined as follows:
# Given a bit string of length $n$, $x \in \{0, 1\}^\otimes n$, we have a function that is either constant or balanced.
#
# A constant function is defined as $f_\text{const}(x) = c \forall x$, where $c \in \{0, 1\}$ is some constant value.
#
# A balanced function, on the other hand, is defined by
#
# $f_\text{balanced}(x) = \begin{cases} 0 \, \forall x \in S(x), \\ 1 \text{ else,} \end{cases}$
#
# where $S(x)$ is an arbitrarily chosen half of all possible bit strings, i.e. $|S(x)| = 2^{n-1}$.


# %% [markdown]
# For our example, we will be using $n + 1$ qubits, where $n$ store the bitstring $x$ and the result is stored in the last qubit.
# We'll be writing our oracle functions as squin kernels, which we can then later use in the actual algorithm implementation.
#
# In order to define our oracle functions, we can simply choose for the constant function to always return $1$, which we achieve by flipping the final qubit using an $X$ gate.
# %%
@squin.kernel
def f_constant(q: ilist.IList[Qubit, Any]):
    # flip the final (result) qubit -- every bit string is mapped to 1
    squin.x(q[-1])


# %% [markdown]

# For the balanced oracle we use the following approach: we use the first qubit as control in a $CX$ gate, which is applied to the resulting qubit.
# This means that the result will be $1$ in exactly half the cases.


# %%
@squin.kernel
def f_balanced(q: ilist.IList[Qubit, Any]):
    squin.cx(q[0], q[-1])


# %% [markdown]
#
# Now, we define the actual algorithm as a kernel, which simply takes one of the other kernels as input.
# In the end, we can infer which function was provided by looking at the resulting measurement of the result qubit.
# %%
@squin.kernel
def deutsch_algorithm(f):
    q = squin.qalloc(n_qubits=n_bits + 1)
    squin.x(q[-1])

    # broadcast for parallelism
    squin.broadcast.h(q)

    # apply the oracle function
    f(q)

    squin.broadcast.h(q[:-1])

    return squin.broadcast.measure(q[:-1])


# %% [markdown]
# Finally, we actually run the result.
# To do so, we use the `PyQrack` simulation backend in bloqade.
#
# To make things a bit more interesting, we randomly select which function we are running the algorithm with.

# %%
sim = StackMemorySimulator(min_qubits=n_bits + 1)

f_choice_idx = random.randint(0, 1)
f_choice = (f_constant, f_balanced)[f_choice_idx]

# result = sim.run(deutsch_algorithm, args=(f_balanced, n))
result0 = 0.0
n_shots = 100
for _ in range(n_shots):
    res = sim.run(deutsch_algorithm, args=(f_choice,))
    result0 += res[0] / n_shots

print(
    "Oh magic Deutsch-Jozsa algorithm, tell us if our function is constant or balanced:"
)
print("*drumroll*")
if result0 == 0:
    print("It's constant!")

    # let's make sure we actually did the right thing here
    assert f_choice_idx == 0
else:
    print("It's balanced!")

    # let's make sure we actually did the right thing here
    assert f_choice_idx == 1
